#include <stdio.h>
#include <time.h>
#include <string.h>

#include <gtk/gtk.h>

#define GTK_EVENT_ACTIVATE "activate"
#define GTK_EVENT_DESTROY "destroy"
#define GTK_EVENT_CLICKED "clicked"
#define GTK_EVENT_CHANGED "changed"

#define DEBUG

/* GObjects */
static GObject *entry_date;
static GObject *entry_hour;
static GObject *entry_minute;
static GObject *entry_second;

/* Time info buffer */
static time_t time_now;
static struct tm *tm_now;
static char active_date[11];
static char active_hour[3];
static char active_min[3];
static char active_sec[3];

/* Timezone info buffer */
#define MAX_TIME_STATUS_LINE 8
static char timezone_cur[32] = {0};
static char timezone_new[32] = {0};
static char timezone_buffer[512][32];
static int timezone_num = 0;

static int get_timezone(const char *src, char *timezone) {
    const char *buf = src;
    int index = 0;

    /* Skip space char */
    while(buf[index] == ' ') index++;

    for (int i = 0; ; i++, index++) {
        if (buf[index] != ' ') {
            timezone[i] = buf[index];
        } else {
            /* End by space */
            timezone[i] = 0;
            break;
        }
    }

    return 0;
}

static void set_timedatectl(char *command, char *data) {
    char cmd[64] = "timedatectl ";
    strcat(cmd, command);
    strcat(cmd, " ");
    strcat(cmd, data);
#ifdef DEBUG
    g_print("[DEBUG] CMD: %s\n", cmd);
#endif

    FILE *fp = popen(cmd, "r");
    pclose(fp);
}

static void set_timezone(char *timezone) {
    if (strcmp(timezone_cur, timezone)) {
        set_timedatectl("set-timezone", timezone);
    }
}

static void set_date_time(const char *date, const char *hour, const char *min, const char *sec) {
    char data[32] = {0};

    if (strcmp(active_date, date)) {
        sprintf(data, "'%s %s:%s:%s'", date, hour, min, sec);
        set_timedatectl("set-time", data);
        strcpy(active_date, date);
        strcpy(active_hour, hour);
        strcpy(active_min, min);
        strcpy(active_sec, sec);
    } else if (strcmp(active_hour, hour) || strcmp(active_min, min) || strcmp(active_sec, sec)) {
        sprintf(data, "'%s:%s:%s'", hour, min, sec);
        set_timedatectl("set-time", data);
        strcpy(active_hour, hour);
        strcpy(active_min, min);
        strcpy(active_sec, sec);
    }
}

static void get_time_date_status() {
    char time_status[8][32];
    FILE *fp = popen("timedatectl status | grep \"Time zone\"", "r");

    for (int i = 0; i < MAX_TIME_STATUS_LINE; i++) {
        char *buf = fgets(time_status[i], 32, fp);
        if (buf == NULL) {
            break;
        } else if (strstr(time_status[i], "Time zone") != NULL) {
            int line_size = strlen(buf);
            fseek(fp, line_size, SEEK_CUR);

            buf = strchr(time_status[i], ':');
            if (buf != NULL) {
                get_timezone(buf + 1, timezone_cur);
                strcpy(timezone_new, timezone_cur);
            }
        }
    }
#ifdef DEBUG
    g_print("[DEBUG] Current timezone: %s\n", timezone_cur);
#endif
    pclose(fp);
}

static void combo_handler(GObject *combo, gpointer user_data) {
    GtkTreeIter  iter;
    GtkTreeModel *model;
    gchar        *string = NULL;

    if (!g_strcmp0((gchar *)user_data, "timezone")) {
        /* Obtain currently selected item from combo box.
         * If nothing is selected, do nothing. */
        if (gtk_combo_box_get_active_iter(GTK_COMBO_BOX(combo), &iter)) {
            /* Obtain data model from combo box. */
            model = gtk_combo_box_get_model(GTK_COMBO_BOX(combo));

            /* Obtain string from model. */
            gtk_tree_model_get(model, &iter, 0, &string, -1);
        }

#ifdef DEBUG
        /* Print string to the console - if string is NULL, print NULL. */
        g_print("[DEBUG] Selected (complex): >> %s <<\n", ( string ? string : "NULL" ));
#endif

        /* Update current timezone */
        strcpy(timezone_new, string);

        /* Free string (if not NULL). */
        if (string)
            g_free(string);
    }
}

static void timezone_init(GtkBuilder *builder) {
    GObject *combo_timezone;
    GtkListStore *store = NULL;
    GtkTreeIter iter;
    GtkTreeIter active_iter;
    GtkCellRenderer *cell = NULL;

    combo_timezone = gtk_builder_get_object(builder, "combobox-timezone");
    if (combo_timezone == NULL) {
        g_print("[ERROR] can not get combobox timezone object\n");
        return;
    }

    FILE *fp = popen("timedatectl list-timezones", "r");

    timezone_num = 0;
    for (int i = 0; i < 512; i++) {
        char *buf = fgets(timezone_buffer[i], 32, fp);
        if (buf == NULL) {
            break;
        } else {
            int size = strlen(timezone_buffer[i]);
            timezone_buffer[i][size - 1] = 0; // remove '\n' char.
            fseek(fp, strlen(buf), SEEK_CUR);
            timezone_num++;
        }
    }

    pclose(fp);

#ifdef DEBUG
    g_print("[DEBUG] Get %d timezones\n", timezone_num);
#endif

    store = gtk_list_store_new(1, G_TYPE_STRING);

    for (int i = 0; i < timezone_num; i++) {
        gtk_list_store_append(store, &iter);
        if (!g_strcmp0(timezone_buffer[i], timezone_cur)) {
            active_iter = iter;
        }
        gtk_list_store_set(store, &iter, 0, timezone_buffer[i], -1);
    }

    gtk_combo_box_set_model(GTK_COMBO_BOX(combo_timezone), GTK_TREE_MODEL(store));
    
    cell = gtk_cell_renderer_text_new();
    gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(combo_timezone), cell, TRUE);
    gtk_cell_layout_set_attributes(GTK_CELL_LAYOUT(combo_timezone), cell, "text", 0, NULL);

    /* Set default active item by iter */
    gtk_combo_box_set_active_iter(GTK_COMBO_BOX(combo_timezone), &active_iter);

    g_signal_connect(combo_timezone, GTK_EVENT_CHANGED, G_CALLBACK(combo_handler), "timezone");
}

static void combobox_timezone_changed_cb(GObject *object, gpointer user_data) {
#ifdef DEBUG
    g_print("combobox_timezone_changed_cb");
#endif
}

static void button_handler(GObject *button, gpointer user_data) {
    const gchar *date;
    const gchar *hour;
    const gchar *min;
    const gchar *sec;

    if (!g_strcmp0((gchar *)user_data, "ok")) {
        date = gtk_entry_get_text(GTK_ENTRY(entry_date));
        hour = gtk_entry_get_text(GTK_ENTRY(entry_hour));
        min = gtk_entry_get_text(GTK_ENTRY(entry_minute));
        sec = gtk_entry_get_text(GTK_ENTRY(entry_second));
        set_date_time(date, hour, min, sec);
        set_timezone(timezone_new);
    } else if (!g_strcmp0((gchar *)user_data, "cancel")) {
        gtk_main_quit();
    }
}

static void date_init(GtkEntry *entry_date, GtkEntry *entry_hour, GtkEntry *entry_minute, GtkEntry *entry_second) {
    /* Get current local time */
    time(&time_now);
    tm_now = localtime(&time_now);

    sprintf(active_date, "%04d-%02d-%02d", (1900 + tm_now->tm_year), (tm_now->tm_mon + 1), tm_now->tm_mday);
#ifdef DEBUG
    g_print("[DEBUG] Date: %s\n", active_date);
#endif

    sprintf(active_hour, "%02d", tm_now->tm_hour);
    sprintf(active_min, "%02d", tm_now->tm_min);
    sprintf(active_sec, "%02d", tm_now->tm_sec);
#ifdef DEBUG
    g_print("[DEBUG] Time: %s:%s:%s\n", active_hour, active_min, active_sec);
#endif

    gtk_entry_set_text(entry_date, active_date);
    gtk_entry_set_text(entry_hour, active_hour);
    gtk_entry_set_text(entry_minute, active_min);
    gtk_entry_set_text(entry_second, active_sec);
}

static int gobjects_init(GtkBuilder *builder, GtkWidget *dialog) {
    GObject *combo_hour;
    GObject *combo_minute;
    GObject *combo_second;
    GObject *button_ok;
    GObject *button_cancel;

    gchar text[3] = {0};

    entry_date = gtk_builder_get_object(builder, "entry-date");
    if (entry_date == NULL) {
        g_print("[ERROR] can not get entry date object\n");
        return -1;
    }

    entry_hour = gtk_builder_get_object(builder, "comboboxtext-entry-hour");
    if (entry_hour == NULL) {
        g_print("[ERROR] can not get entry hour object\n");
        return -1;
    }

    entry_minute = gtk_builder_get_object(builder, "comboboxtext-entry-minute");
    if (entry_minute == NULL) {
        g_print("[ERROR] can not get entry minute object\n");
        return -1;
    }

    entry_second = gtk_builder_get_object(builder, "comboboxtext-entry-second");
    if (entry_second == NULL) {
        g_print("[ERROR] can not get entry second object\n");
        return -1;
    }

    combo_hour = gtk_builder_get_object(builder, "comboboxtext-hour");
    if (combo_hour == NULL) {
        g_print("[ERROR] can not get combo box hour object\n");
        return -1;
    }
    /* Add hour combo box context */
    for (int i = 0; i < 24; i++) {
        sprintf(text, "%d", i);
        gtk_combo_box_text_append(GTK_COMBO_BOX_TEXT(combo_hour), text, text);
    }

    combo_minute = gtk_builder_get_object(builder, "comboboxtext-minute");
    if (combo_minute == NULL) {
        g_print("[ERROR] can not get combo box minute object\n");
        return -1;
    }
    /* Add minute combo box context */
    for (int i = 0; i < 60; i++) {
        sprintf(text, "%d", i);
        gtk_combo_box_text_append(GTK_COMBO_BOX_TEXT(combo_minute), text, text);
    }

    combo_second = gtk_builder_get_object(builder, "comboboxtext-second");
    if (combo_second == NULL) {
        g_print("[ERROR] can not get combo box second object\n");
        return -1;
    }
    /* Add second combo box context */
    for (int i = 0; i < 60; i++) {
        sprintf(text, "%d", i);
        gtk_combo_box_text_append(GTK_COMBO_BOX_TEXT(combo_second), text, text);
    }

    button_ok = gtk_builder_get_object(builder, "button-ok");
    if (button_ok == NULL) {
        g_print("[ERROR] can not get button ok object\n");
        return -1;
    }

    button_cancel = gtk_builder_get_object(builder, "button-cancel");
    if (button_cancel == NULL) {
        g_print("[ERROR] can not get button cancel object\n");
        return -1;
    }

    date_init(GTK_ENTRY(entry_date), GTK_ENTRY(entry_hour), GTK_ENTRY(entry_minute), GTK_ENTRY(entry_second));

    // /* Set default time combo box item */
    // gtk_combo_box_set_active_id(GTK_COMBO_BOX(combo_hour), active_hour);
    // gtk_combo_box_set_active_id(GTK_COMBO_BOX(combo_minute), active_min);
    // gtk_combo_box_set_active_id(GTK_COMBO_BOX(combo_second), active_sec);
    // gtk_combo_box_set_active(GTK_COMBO_BOX(combo_hour), tm_now->tm_hour);

    // g_signal_connect(entry_date, GTK_EVENT_ACTIVATE, G_CALLBACK(date_activate), "date");
    g_signal_connect(combo_hour, GTK_EVENT_CHANGED, G_CALLBACK(combo_handler), "hour");
    g_signal_connect(button_ok, GTK_EVENT_CLICKED, G_CALLBACK(button_handler), "ok");
    g_signal_connect(button_cancel, GTK_EVENT_CLICKED, G_CALLBACK(button_handler), "cancel");

    return 0;
}

int main(int argc, char **argv) {
    GtkBuilder *builder = gtk_builder_new();
    GtkWidget *dialog;

    gtk_init(&argc, &argv);

    if (!gtk_builder_add_from_file(builder, "timemanager.ui", NULL)) {
        g_print("[ERROR] can not load glade file\n");
        return -1;
    }

    dialog = GTK_WIDGET(gtk_builder_get_object(builder, "dialog-date_time"));
    if (dialog == NULL) {
        g_print("[ERROR] can not get dialog object\n");
        return -1;
    }

    gtk_window_set_title(GTK_WINDOW(dialog), "Date & Time");
    gtk_window_set_default_size(GTK_WINDOW(dialog), 400, 200);

    gtk_builder_connect_signals(builder, NULL);
    g_signal_connect(dialog, GTK_EVENT_DESTROY, G_CALLBACK(gtk_main_quit), NULL);

    gobjects_init(builder, dialog);

    get_time_date_status();
    timezone_init(builder);

    g_object_unref(G_OBJECT (builder));
    gtk_widget_show_all(dialog);
    gtk_main();

    return 0;
}
